# Generated from D:/Documents/pycharm_workspace/Qrogue/qrogue/game/world/dungeon_generator\QrogueDungeon.g4 by ANTLR 4.9.2
from antlr4 import *

# This class defines a complete listener for a parse tree produced by QrogueDungeonParser.
class QrogueDungeonListener(ParseTreeListener):

    # Enter a parse tree produced by QrogueDungeonParser#start.
    def enterStart(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#start.
    def exitStart(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#integer.
    def enterInteger(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#integer.
    def exitInteger(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#complex_number.
    def enterComplex_number(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#complex_number.
    def exitComplex_number(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#robot.
    def enterRobot(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#robot.
    def exitRobot(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#layout.
    def enterLayout(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#layout.
    def exitLayout(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#l_room_row.
    def enterL_room_row(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#l_room_row.
    def exitL_room_row(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#l_hallway_row.
    def enterL_hallway_row(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#l_hallway_row.
    def exitL_hallway_row(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#rooms.
    def enterRooms(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#rooms.
    def exitRooms(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#room.
    def enterRoom(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#room.
    def exitRoom(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_attributes.
    def enterR_attributes(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_attributes.
    def exitR_attributes(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_visibility.
    def enterR_visibility(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_visibility.
    def exitR_visibility(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_type.
    def enterR_type(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_type.
    def exitR_type(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_row.
    def enterR_row(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_row.
    def exitR_row(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#tile.
    def enterTile(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#tile.
    def exitTile(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#tile_descriptor.
    def enterTile_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#tile_descriptor.
    def exitTile_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#trigger_descriptor.
    def enterTrigger_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#trigger_descriptor.
    def exitTrigger_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#message_descriptor.
    def enterMessage_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#message_descriptor.
    def exitMessage_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#enemy_descriptor.
    def enterEnemy_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#enemy_descriptor.
    def exitEnemy_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#collectible_descriptor.
    def enterCollectible_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#collectible_descriptor.
    def exitCollectible_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#energy_descriptor.
    def enterEnergy_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#energy_descriptor.
    def exitEnergy_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#riddle_descriptor.
    def enterRiddle_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#riddle_descriptor.
    def exitRiddle_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#shop_descriptor.
    def enterShop_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#shop_descriptor.
    def exitShop_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#hallways.
    def enterHallways(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#hallways.
    def exitHallways(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#hallway.
    def enterHallway(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#hallway.
    def exitHallway(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#h_attributes.
    def enterH_attributes(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#h_attributes.
    def exitH_attributes(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#draw_strategy.
    def enterDraw_strategy(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#draw_strategy.
    def exitDraw_strategy(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stv_pools.
    def enterStv_pools(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stv_pools.
    def exitStv_pools(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#default_stv_pool.
    def enterDefault_stv_pool(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#default_stv_pool.
    def exitDefault_stv_pool(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stv_pool.
    def enterStv_pool(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stv_pool.
    def exitStv_pool(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stvs.
    def enterStvs(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stvs.
    def exitStvs(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stv.
    def enterStv(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stv.
    def exitStv(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#reward_pools.
    def enterReward_pools(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#reward_pools.
    def exitReward_pools(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#default_reward_pool.
    def enterDefault_reward_pool(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#default_reward_pool.
    def exitDefault_reward_pool(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#reward_pool.
    def enterReward_pool(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#reward_pool.
    def exitReward_pool(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#collectibles.
    def enterCollectibles(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#collectibles.
    def exitCollectibles(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#collectible.
    def enterCollectible(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#collectible.
    def exitCollectible(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#messages.
    def enterMessages(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#messages.
    def exitMessages(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#message.
    def enterMessage(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#message.
    def exitMessage(self, ctx):
        pass


