# Generated from D:/Documents/pycharm_workspace/Qrogue/qrogue/dungeon_editor\QrogueWorld.g4 by ANTLR 4.9.2
from antlr4 import *

# This class defines a complete listener for a parse tree produced by QrogueWorldParser.
class QrogueWorldListener(ParseTreeListener):

    # Enter a parse tree produced by QrogueWorldParser#start.
    def enterStart(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#start.
    def exitStart(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#layout.
    def enterLayout(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#layout.
    def exitLayout(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#l_room_row.
    def enterL_room_row(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#l_room_row.
    def exitL_room_row(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#l_hallway_row.
    def enterL_hallway_row(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#l_hallway_row.
    def exitL_hallway_row(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#rooms.
    def enterRooms(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#rooms.
    def exitRooms(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#room.
    def enterRoom(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#room.
    def exitRoom(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#r_attributes.
    def enterR_attributes(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#r_attributes.
    def exitR_attributes(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#r_visibility.
    def enterR_visibility(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#r_visibility.
    def exitR_visibility(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#r_type.
    def enterR_type(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#r_type.
    def exitR_type(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#hallways.
    def enterHallways(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#hallways.
    def exitHallways(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#hallway.
    def enterHallway(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#hallway.
    def exitHallway(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#h_attributes.
    def enterH_attributes(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#h_attributes.
    def exitH_attributes(self, ctx):
        pass


